/*****************************************************************************
*
* Copyright (c) 2000 - 2013, Lawrence Livermore National Security, LLC
* Produced at the Lawrence Livermore National Laboratory
* LLNL-CODE-442911
* All rights reserved.
*
* This file is  part of VisIt. For  details, see https://visit.llnl.gov/.  The
* full copyright notice is contained in the file COPYRIGHT located at the root
* of the VisIt distribution or at http://www.llnl.gov/visit/copyright.html.
*
* Redistribution  and  use  in  source  and  binary  forms,  with  or  without
* modification, are permitted provided that the following conditions are met:
*
*  - Redistributions of  source code must  retain the above  copyright notice,
*    this list of conditions and the disclaimer below.
*  - Redistributions in binary form must reproduce the above copyright notice,
*    this  list of  conditions  and  the  disclaimer (as noted below)  in  the
*    documentation and/or other materials provided with the distribution.
*  - Neither the name of  the LLNS/LLNL nor the names of  its contributors may
*    be used to endorse or promote products derived from this software without
*    specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT  HOLDERS AND CONTRIBUTORS "AS IS"
* AND ANY EXPRESS OR  IMPLIED WARRANTIES, INCLUDING,  BUT NOT  LIMITED TO, THE
* IMPLIED WARRANTIES OF MERCHANTABILITY AND  FITNESS FOR A PARTICULAR  PURPOSE
* ARE  DISCLAIMED. IN  NO EVENT  SHALL LAWRENCE  LIVERMORE NATIONAL  SECURITY,
* LLC, THE  U.S.  DEPARTMENT OF  ENERGY  OR  CONTRIBUTORS BE  LIABLE  FOR  ANY
* DIRECT,  INDIRECT,   INCIDENTAL,   SPECIAL,   EXEMPLARY,  OR   CONSEQUENTIAL
* DAMAGES (INCLUDING, BUT NOT  LIMITED TO, PROCUREMENT OF  SUBSTITUTE GOODS OR
* SERVICES; LOSS OF  USE, DATA, OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER
* CAUSED  AND  ON  ANY  THEORY  OF  LIABILITY,  WHETHER  IN  CONTRACT,  STRICT
* LIABILITY, OR TORT  (INCLUDING NEGLIGENCE OR OTHERWISE)  ARISING IN ANY  WAY
* OUT OF THE  USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
* DAMAGE.
*
*****************************************************************************/

#include <ExodusPluginInfo.h>
#include <avtExodusFileFormat.h>
#include <avtMTSDFileFormatInterface.h>
#include <avtGenericDatabase.h>

// ****************************************************************************
//  Method:  ExodusCommonPluginInfo::GetDatabaseType
//
//  Purpose:
//    Returns the type of a Exodus database.
//
//  Programmer:  generated by xml2info
//  Creation:    omitted
//
// ****************************************************************************
DatabaseType
ExodusCommonPluginInfo::GetDatabaseType()
{
    return DB_TYPE_MTSD;
}

// ****************************************************************************
//  Method: ExodusCommonPluginInfo::SetupDatabase
//
//  Purpose:
//      Sets up a Exodus database.
//
//  Arguments:
//      list    A list of file names.
//      nList   The number of timesteps in list.
//      nBlocks The number of blocks in the list.
//
//  Returns:    A Exodus database from list.
//
//  Programmer: Jeremy Meredith
//  Creation:   August 27, 2002
//
//  Modifications:
//
//    Hank Childs, Tue Feb 25 13:56:42 PST 2003
//    Account for .EXODUS, .nemesis, and .NEMESIS.
//
//    Hank Childs, Thu Jul 22 11:29:31 PDT 2004
//    Register the file list 
//
//    Jeremy Meredith, Thu Jan 28 12:28:07 EST 2010
//    MTSD now accepts grouping multiple files into longer sequences, so
//    its interface has changed to accept both a number of timestep groups
//    and a number of blocks.
//
//    Brad Whitlock, Tue Feb 2 10:06:34 PDT 2010
//    I changed the code so when nBlocks==-1 it will treat the list of files
//    as domains so multiblock files work again.
//
//    Mark C. Miller, Wed Jan  8 18:07:37 PST 2014
//    I enabled this to work for cases in which we have timestep groups *and*
//    many files as well as the most common case where nBlocks is number of
//    files in the list.
// ****************************************************************************
#include <string>

avtDatabase *
ExodusCommonPluginInfo::SetupDatabase(const char *const *list,
                                   int nList, int nBlock)
{
    std::string file1 = list[0];

    bool containsManyFiles = false;
    if (nList == 1 &&
        file1.length() > 7 && file1.substr(file1.length()-7,7)==".exodus")
        containsManyFiles = true;
    if (nList == 1 &&
        file1.length() > 7 && file1.substr(file1.length()-7,7)==".EXODUS")
        containsManyFiles = true;
    if (nList == 1 &&
        file1.length() > 8 && file1.substr(file1.length()-8,8)==".nemesis")
        containsManyFiles = true;
    if (nList == 1 &&
        file1.length() > 8 && file1.substr(file1.length()-8,8)==".NEMESIS")
        containsManyFiles = true;

    //
    // Note that in the context of Exodus, the !NBLOCKS directive and 'nBlocks' here
    // is *not* equivalent to Exodus' notion of 'Element Blocks'. Instead, it is really
    // the number of independent files (or series of files for cases where the output is
    // broken into multiple timestep groups).
    //

    if (containsManyFiles)
    {
        const char *filename = list[0];
        char  **reallist  = NULL;
        int     listcount = 0;
        int     bang_nBlocks=-1;
        avtDatabase::GetFileListFromTextFile(filename, reallist, listcount, &bang_nBlocks);

        avtDatabase *rv = ExodusCommonPluginInfo::SetupDatabase(reallist,
                              listcount, bang_nBlocks);

        //
        // Clean up memory
        //
        for (int i = 0 ; i < listcount ; i++)
        {
            delete [] reallist[i];
        }
        delete [] reallist;

        return rv;
    }

    //
    // We don't want to register the file list with every Exodus file format,
    // because that list can get big.  Instead, register a list statically
    // with the format.  It will return an index and then tell each new
    // instance that it should use this index.
    //
    int fileListId = -1;
    if (!containsManyFiles)
    {
        fileListId = avtExodusFileFormat::RegisterFileList(list, nList);
        nBlock = nList;
    }

    int nTimestepGroups = 1;
    if(nBlock == -1)
    {
        // If nBlock == -1 then we're in here recursively and we're dealing
        // with a list of domains.
        nBlock = nList;
    }

    nTimestepGroups = nList / nBlock;

    avtMTSDFileFormat ***ffl = new avtMTSDFileFormat**[nTimestepGroups];
    for (int i = 0 ; i < nTimestepGroups ; i++)
    {
        ffl[i] = new avtMTSDFileFormat*[nBlock];
        for (int j = 0 ; j < nBlock ; j++)
        {
            avtExodusFileFormat *exo = new avtExodusFileFormat(list[i*nBlock+j]);
            if (!containsManyFiles)
                exo->SetFileList(fileListId);
            ffl[i][j] = exo; 
        }
    }
    avtMTSDFileFormatInterface *inter 
        = new avtMTSDFileFormatInterface(ffl, nTimestepGroups, nBlock);
    return new avtGenericDatabase(inter);
}

